from cyclonedx.model import (
    HashType,
    Property,
    XsUri,
)
from cyclonedx.model.contact import (
    OrganizationalContact,
)
from cyclonedx.model.vulnerability import (
    BomTarget,
    BomTargetVersionRange,
    Vulnerability,
    VulnerabilityAdvisory,
    VulnerabilityRating,
    VulnerabilitySeverity,
)
from decimal import (
    Decimal,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.pkg.package import (
    HealthMetadata,
    Package,
)
from fluid_sbom.utils.package import (
    get_author_info,
)

# Constants
NAMESPACE = "fluid-attacks"


def add_vulnerabilities(package: Package) -> list[Vulnerability]:
    vulnerabilities = []
    if package.advisories:
        for advisory in package.advisories:
            rating = VulnerabilityRating(
                severity=VulnerabilitySeverity(advisory.severity.lower()),
                score=Decimal(str(advisory.percentile)),
            )

            advisory_obj = (
                [
                    VulnerabilityAdvisory(url=XsUri(url))
                    for url in advisory.urls
                ]
                if advisory.urls
                else []
            )

            bom_target = BomTarget(
                ref=f"{package.name}@{package.version}",
                versions=[
                    BomTargetVersionRange(
                        version=package.version
                        if not advisory.version_constraint
                        else None,
                        range=advisory.version_constraint,
                    )
                ],
            )

            additional_properties = [
                Property(
                    name=f"{NAMESPACE}:advisory:epss", value=str(advisory.epss)
                )
            ]
            vulnerability = Vulnerability(
                bom_ref=f"{package.name}@{advisory.id}",
                id=advisory.id,
                description=advisory.description,
                advisories=advisory_obj,
                ratings=[rating],
                affects=[bom_target],
                properties=additional_properties,
            )
            vulnerabilities.append(vulnerability)

    return vulnerabilities


def add_health_metadata_properties(
    health_metadata: HealthMetadata,
) -> list[Property]:
    properties = []
    if health_metadata.latest_version:
        properties.append(
            Property(
                name=f"{NAMESPACE}:health_metadata:latest_version",
                value=health_metadata.latest_version,
            )
        )

    if health_metadata.latest_version_created_at:
        properties.append(
            Property(
                name=f"{NAMESPACE}:health_metadata:latest_version_created_at",
                value=str(health_metadata.latest_version_created_at),
            )
        )

    return properties


def add_locations_properties(locations: list[Location]) -> list[Property]:
    properties = []
    for idx, location in enumerate(locations):
        path = location.path()
        line = (
            location.coordinates.line
            if location.coordinates and location.coordinates.line
            else None
        )
        layer = (
            location.coordinates.file_system_id
            if location.coordinates and location.coordinates.file_system_id
            else None
        )

        properties.append(
            Property(name=f"{NAMESPACE}:locations:{idx}:path", value=path)
        )
        if line:
            properties.append(
                Property(
                    name=f"{NAMESPACE}:locations:{idx}:line", value=str(line)
                )
            )
        if layer:
            properties.append(
                Property(
                    name=f"{NAMESPACE}:locations:{idx}:layer", value=layer
                )
            )

    return properties


def add_component_properties(package: Package) -> list[Property]:
    properties = [
        Property(name=f"{NAMESPACE}:language", value=package.language)
    ]

    if package.health_metadata:
        properties.extend(
            add_health_metadata_properties(package.health_metadata)
        )

    if package.locations:
        properties.extend(add_locations_properties(package.locations))

    return properties


def add_integrity(health_metadata: HealthMetadata) -> list[HashType] | None:
    integrity = (
        health_metadata.artifact.integrity
        if health_metadata.artifact
        else None
    )
    if integrity and integrity.algorithm and integrity.value:
        return [
            HashType.from_hashlib_alg(integrity.algorithm, integrity.value)
        ]
    return None


def add_authors(
    health_metadata: HealthMetadata | None,
) -> list[OrganizationalContact]:
    if health_metadata is None:
        return []

    authors_info = get_author_info(health_metadata)
    authors = []

    for name, email in authors_info:
        if name or email:
            authors.append(
                OrganizationalContact(
                    name=name,
                    email=email,
                )
            )

    return authors



from fluid_sbom.internal.cache import (
    dual_cache,
)
import requests
from typing import (
    Any,
)


@dual_cache
def get_composer_package(
    package_name: str, version: str | None = None
) -> dict[str, Any] | None:
    base_url = f"https://repo.packagist.org/p2/{package_name}.json"
    response = requests.get(base_url, timeout=20)
    if response.status_code == 200:
        response_data = response.json()

        package_versions = response_data.get("packages", {}).get(
            package_name, []
        )

        if version:
            for version_data in package_versions:
                if version_data.get("version") == version:
                    return version_data
            return None

        if package_versions:
            latest_version_data = package_versions[0]
            return latest_version_data

        return None

    return None



from fluid_sbom.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection import (
    toml,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.python.package import (
    package_url,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.python import (
    PythonRequirementsEntry,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from fluid_sbom.utils.exceptions import (
    UnexpectedNode,
)
import logging
from pydantic import (
    ValidationError,
)

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def parse_poetry_lock(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    _content = reader.read_closer.read()

    try:
        toml_content: IndexedDict = toml.parse_toml_with_tree_sitter(_content)
    except UnexpectedNode:
        return [], []

    if not toml_content:
        return [], []

    packages = _parse_packages(toml_content, reader)
    relationships = _parse_relationships(toml_content, packages)

    return packages, relationships


def _parse_packages(
    toml_content: IndexedDict, reader: LocationReadCloser
) -> list[Package]:
    packages = []

    for package in toml_content.get("package", []):
        name: str | None = package.get("name")
        version: str | None = package.get("version")

        if not name or not version:
            continue

        p_url = package_url(name, version, package)

        location = (
            _get_location(reader.location, package.position.start.line)
            if isinstance(package, IndexedDict)
            else reader.location
        )

        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    found_by=None,
                    locations=[location],
                    language=Language.PYTHON,
                    p_url=p_url,
                    metadata=PythonRequirementsEntry(
                        name=name,
                        extras=[],
                        markers=p_url,
                    ),
                    licenses=[],
                    type=PackageType.PythonPkg,
                )
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types "
                "are incorrect.",
                extra={
                    "extra": {
                        "exception": ex.errors(include_url=False),
                        "location": location.path(),
                    }
                },
            )
            continue

    return packages


def _parse_relationships(
    toml_content: IndexedDict,
    packages: list[Package],
) -> list[Relationship]:
    relationships = []

    for package in toml_content.get("package", []):
        _pkg = next(
            (pkg for pkg in packages if pkg.name == package["name"]), None
        )
        dependencies = list(package.get("dependencies", {}).keys())

        if _pkg and dependencies:
            for dep in dependencies:
                dep_pkg = next(
                    (pkg for pkg in packages if pkg.name == dep), None
                )
                if dep_pkg:
                    relationships.append(
                        Relationship(
                            from_=dep_pkg,
                            to_=_pkg,
                            type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                            data=None,
                        )
                    )

    return relationships



# pylint:disable=no-value-for-parameter
import click
from concurrent.futures import (
    ThreadPoolExecutor,
)
from fluid_sbom.advisories.database import (
    DATABASE,
)
from fluid_sbom.cli.options import (
    MutuallyExclusiveOption,
    RequiredAnyCommand,
)
from fluid_sbom.cli.sbom_config import (
    build_sbom_config,
    ScanArgs,
)
from fluid_sbom.config.bugsnag import (
    initialize_bugsnag,
)
from fluid_sbom.config.config import (
    SourceType,
)
from fluid_sbom.config.logger import (
    configure_logger,
)
from fluid_sbom.context.image import (
    get_context,
)
from fluid_sbom.format import (
    format_sbom,
)
from fluid_sbom.internal.file_resolver.container_image import (
    ContainerImage,
)
from fluid_sbom.internal.file_resolver.directory import (
    Directory,
)
from fluid_sbom.internal.operations.package_operation import (
    package_operations_factory,
)
from fluid_sbom.pkg.cataloger.complete import (
    complete_package,
)
from fluid_sbom.sources.docker import (
    get_docker_image,
)
import logging
import os
import textwrap
from typing import (
    Unpack,
)

LOGGER = logging.getLogger(__name__)


def show_banner() -> None:
    logo = textwrap.dedent(
        """
         ───── ⌝
        |    ⌝|  Fluid Attacks
        |  ⌝  |  We hack your software.
         ─────
        """
    )
    click.secho(logo, fg="red")


@click.command(cls=RequiredAnyCommand, required_any=["o_from", "config"])
@click.argument("arg")
@click.option(
    "--from",
    "o_from",
    type=click.Choice(
        ["docker", "dir", "docker-daemon"], case_sensitive=False
    ),
    help=(
        "Source of the scan: 'docker' for scanning Docker images "
        "or 'dir' for scanning directories."
    ),
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["config"],
)
@click.option(
    "--format",
    "output_format",
    type=click.Choice(
        [
            "fluid-json",
            "cyclonedx-json",
            "spdx-json",
            "cyclonedx-xml",
            "spdx-xml",
        ],
        case_sensitive=False,
    ),
    default="fluid-json",
    help="Output format for the scanned data.",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["config"],
)
@click.option(
    "--output",
    "-o",
    help="Output filename.",
    default="sbom",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["config"],
)
@click.option(
    "--docker-user",
    default=None,
    help="Docker registry username.",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["config"],
)
@click.option(
    "--docker-password",
    default=None,
    help="Docker registry password.",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["config"],
)
@click.option(
    "--config",
    "-c",
    is_flag=True,
    default=False,
    cls=MutuallyExclusiveOption,
    help="Path to an advanced configuration file with additional settings.",
    mutually_exclusive=[
        "docker-user",
        "docker-password",
        "output_format",
        "output",
        "o_from",
    ],
)
def scan(arg: str, **kwargs: Unpack[ScanArgs]) -> None:
    configure_logger()
    initialize_bugsnag()
    show_banner()

    sbom_config = build_sbom_config(arg, **kwargs)

    match sbom_config.source_type:
        case SourceType.DIRECTORY:
            sbom_config.resolver = Directory(
                root=sbom_config.source, exclude=sbom_config.exclude
            )
        case SourceType.DOCKER | SourceType.DOCKER_DAEMON:
            daemon = sbom_config.source_type == SourceType.DOCKER_DAEMON
            docker_image = get_docker_image(
                sbom_config.source,
                username=sbom_config.docker_user,
                password=sbom_config.docker_password,
                daemon=daemon,
            )
            if not docker_image:
                raise ValueError(f"No image found for {sbom_config.source}")

            context = get_context(
                docker_image,
                username=sbom_config.docker_user,
                password=sbom_config.docker_password,
                daemon=daemon,
            )
            if context is None:
                raise ValueError(f"No context found for {docker_image}")
            sbom_config.resolver = ContainerImage(
                img=docker_image, context=context, lazy=False
            )
        case _:
            raise ValueError(f"Unknown source: {sbom_config.source}")

    DATABASE.initialize()
    LOGGER.info(
        "📦 Generating SBOM from %s: %s",
        sbom_config.source_type.value,
        sbom_config.source,
    )

    packages, relationships = package_operations_factory(sbom_config.resolver)
    with ThreadPoolExecutor(
        max_workers=min(
            32, (os.cpu_count() or 1) * 5 if os.cpu_count() is not None else 32
        )
    ) as executor:
        LOGGER.info("📦 Gathering additional package information")
        packages = list(filter(None, executor.map(complete_package, packages)))

    LOGGER.info("📦 Preparing %s report", sbom_config.output_format)
    format_sbom(packages, relationships, sbom_config)


if __name__ == "__main__":
    scan(prog_name="sbom")



import confuse
from fluid_sbom.config.config import (
    SbomConfig,
    SbomOutputFormat,
    SourceType,
)
from fluid_sbom.utils.exceptions import (
    InvalidConfigFile,
)
import logging
import os
from typing import (
    TypedDict,
    Unpack,
)

LOGGER = logging.getLogger(__name__)


class ScanArgs(TypedDict):
    o_from: str
    output_format: str
    output: str
    docker_user: str | None
    docker_password: str | None
    config: bool


def _build_config(config_path: str) -> confuse.Configuration:
    template = confuse.Configuration("sbom", read=False)
    template.set_file(config_path)
    template.read(user=False, defaults=False)
    return template.get(
        confuse.Template(
            {
                "source": confuse.String(),
                "source_type": confuse.String(),
                "exclude": confuse.Sequence(confuse.String()),
                "docker_credentials": confuse.Optional(
                    confuse.Template(
                        {
                            "username": confuse.Optional(confuse.String()),
                            "password": confuse.Optional(confuse.String()),
                        }
                    )
                ),
                "output": confuse.Template(
                    {
                        "name": confuse.String(),
                        "format": confuse.OneOf(
                            [_format.value for _format in SbomOutputFormat]
                        ),
                    }
                ),
            }
        ),
    )


def load(config_path: str) -> SbomConfig:
    config = _build_config(config_path)
    config_docker = config.pop("docker_credentials", {})

    try:
        output = config.pop("output", None)

        sbom_config = SbomConfig(
            source=config.pop("source", None),
            source_type=config.pop("source_type", "dir"),
            output_format=output["format"],
            output=output["name"],
            resolver=None,
            exclude=config.pop("exclude", None) or tuple(),
            docker_user=config_docker.pop("username", None),
            docker_password=config_docker.pop("password", None),
        )

        if config:
            unrecognized_keys = ", ".join(config)
            msg = (
                f"Some keys were not recognized: {unrecognized_keys}."
                " The analysis will be performed only using the supported keys"
                " and defaults."
            )
            LOGGER.warning(msg)
    except KeyError as exc:
        raise confuse.ConfigError(f"Key: {exc.args[0]} is required")

    return sbom_config


def build_sbom_config(arg: str, **kwargs: Unpack[ScanArgs]) -> SbomConfig:
    if kwargs["config"]:
        if os.path.isfile(arg):
            if not arg.endswith((".yaml", ".yml")):
                raise InvalidConfigFile(
                    "The configuration file must be a YAML format"
                )
            return load(arg)
        raise InvalidConfigFile(
            f"The configuration file is not a valid file: {arg}"
        )

    return SbomConfig(
        source=arg,
        source_type=SourceType.from_string(kwargs["o_from"]),
        output_format=kwargs["output_format"],
        output=kwargs["output"],
        resolver=None,
        exclude=(),
        docker_user=kwargs["docker_user"],
        docker_password=kwargs["docker_password"],
    )



from fluid_sbom.internal.package_information.javascript import (
    NPMPackageLicense,
)
from fluid_sbom.pkg.algorithm import (
    algorithm_length,
)
from fluid_sbom.pkg.package import (
    HealthMetadata,
)
import os
import re
from urllib.parse import (
    urlunparse,
)
import uuid


def strip_version_specifier(item: str) -> str:
    # Define the characters that indicate the start of a version specifier
    specifiers = "[(<>="

    # Find the index of the first occurrence of any specifier character
    index = next(
        (i for i, char in enumerate(item) if char in specifiers), None
    )

    # If no specifier character is found, return the original string
    if index is None:
        return item.strip()

    # Return the substring up to the first specifier character, stripped of
    # leading/trailing whitespace
    return item[:index].strip()


def handle_licenses(
    licenses: str | list[str | dict[str, str]] | NPMPackageLicense,
) -> list[str]:
    if isinstance(licenses, dict):
        return [licenses["type"]] if "type" in licenses else []
    if isinstance(licenses, list):
        licenses_list = []
        for license_item in licenses:
            if isinstance(license_item, str):
                licenses_list.append(license_item)
            if isinstance(license_item, dict) and license_item["type"]:
                licenses_list.append(license_item["type"])
        return licenses_list
    return [licenses]


def is_valid_email(email: str) -> bool:
    email_regex = r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"
    return re.match(email_regex, email) is not None


def get_author_info(
    health_metadata: HealthMetadata,
) -> list[tuple[str | None, str | None]]:
    author_info = []

    if health_metadata.authors:
        authors_list = health_metadata.authors.split(", ")
        for author in authors_list:
            name: str | None = None
            email: str | None = None
            email_match = re.search(r"<([^<>]+)>", author)

            if email_match:
                email_candidate: str = email_match.group(1)
                if is_valid_email(email_candidate):
                    email = email_candidate
                else:
                    email = None
                name = author.replace(email_match.group(0), "").strip() or None
            else:
                name = author.strip() or None

            author_info.append((name, email))

    return author_info


def infer_algorithm(digest_value: str | None) -> str | None:
    if digest_value:
        for algorithm, length in algorithm_length.items():
            if len(digest_value) == int(length):
                return algorithm.value
    return None


def sanitize_name(text: str) -> str:
    return re.sub(r"[^a-zA-Z0-9.-]", "-", text)


def get_document_namespace(working_dir: str) -> str:
    input_type = "unknown-source-type"

    if os.path.isfile(working_dir):
        input_type = "file"
    elif os.path.isdir(working_dir):
        input_type = "dir"

    unique_id = uuid.uuid4()
    identifier = os.path.join(input_type, str(unique_id))
    if working_dir != ".":
        identifier = os.path.join(input_type, f"{working_dir}-{unique_id}")

    doc_namespace = urlunparse(
        ("https", "fluidattacks.com", identifier, "", "", ""),
    )

    return doc_namespace
